@@ -39,8 +39,12 @@ allocate_domain_info(domain_table_type *table,
 	result->wildcard_child_closest_match = result;
 	result->rrsets = NULL;
 	result->number = 0;
-#ifdef PLUGINS
-	result->plugin_data = NULL;
+#ifdef NSEC3
+	result->nsec3_exact = NULL;
+	result->nsec3_cover = NULL;
+	result->nsec3_wcard_child_cover = NULL;
+	result->nsec3_ds_parent_exact = NULL;
+	result->nsec3_ds_parent_cover = NULL;
 #endif
 	result->is_existing = 0;
 	result->is_apex = 0;
@@ -64,19 +68,23 @@ domain_table_create(region_type *region)
 	root->parent = NULL;
 	root->wildcard_child_closest_match = root;
 	root->rrsets = NULL;
-	root->number = 0;
-#ifdef PLUGINS
-	root->plugin_data = NULL;
-#endif
+	root->number = 1; /* 0 is used for after header */
 	root->is_existing = 0;
 	root->is_apex = 0;
+#ifdef NSEC3
+	root->nsec3_exact = NULL;
+	root->nsec3_cover = NULL;
+	root->nsec3_wcard_child_cover = NULL;
+	root->nsec3_ds_parent_exact = NULL;
+	root->nsec3_ds_parent_cover = NULL;
+#endif
 	
 	result = (domain_table_type *) region_alloc(region,
 						    sizeof(domain_table_type));
 	result->region = region;
-	result->names_to_domains = heap_create(
+	result->names_to_domains = rbtree_create(
 		region, (int (*)(const void *, const void *)) dname_compare);
-	heap_insert(result->names_to_domains, (rbnode_t *) root);
+	rbtree_insert(result->names_to_domains, (rbnode_t *) root);
 
 	result->root = root;
 
@@ -154,7 +162,8 @@ domain_table_insert(domain_table_type *table,
 			result = allocate_domain_info(table,
 						      dname,
 						      closest_encloser);
-			heap_insert(table->names_to_domains, (rbnode_t *) result);
+			rbtree_insert(table->names_to_domains, (rbnode_t *) result);
+			result->number = table->names_to_domains->count;
 
 			/*
 			 * If the newly added domain name is larger
@@ -189,7 +198,7 @@ domain_table_iterate(domain_table_type *table,
 
 	assert(table);
 
-	HEAP_WALK(table->names_to_domains, dname, node) {
+	RBTREE_WALK(table->names_to_domains, dname, node) {
 		iterator((domain_type *) node, user_data);
 	}
 }
@@ -197,8 +206,17 @@ domain_table_iterate(domain_table_type *table,
 void
 domain_add_rrset(domain_type *domain, rrset_type *rrset)
 {
+#if 0 	/* fast */
 	rrset->next = domain->rrsets;
 	domain->rrsets = rrset;
+#else
+	/* preserve ordering, add at end */
+	rrset_type** p = &domain->rrsets;
+	while(*p) 
+		p = &((*p)->next);
+	*p = rrset;
+	rrset->next = 0;
+#endif
 
 	while (domain && !domain->is_existing) {
 		domain->is_existing = 1;
@@ -335,3 +353,25 @@ namedb_find_zone(namedb_type *db, domain_type *domain)
 
 	return zone;
 }
+
+rrset_type *
+domain_find_non_cname_rrset(domain_type *domain, zone_type *zone)
+{
+	/* find any rrset type that is not allowed next to a CNAME */
+	/* nothing is allowed next to a CNAME, except RRSIG, NSEC, NSEC3 */
+	rrset_type *result = domain->rrsets;
+
+	while (result) {
+		if (result->zone == zone && /* here is the list of exceptions*/
+			rrset_rrtype(result) != TYPE_CNAME &&
+			rrset_rrtype(result) != TYPE_RRSIG &&
+			rrset_rrtype(result) != TYPE_NXT &&
+			rrset_rrtype(result) != TYPE_SIG &&
+			rrset_rrtype(result) != TYPE_NSEC &&
+			rrset_rrtype(result) != TYPE_NSEC3 ) {
+			return result;
+		}
+		result = result->next;
+	}
+	return NULL;
+}
